@@ -1,11 +1,12 @@
 """Asterix data processing module (generic code)
 """
 
-from itertools import chain
+from itertools import chain, dropwhile
 from dataclasses import dataclass
 from binascii import hexlify, unhexlify
 from typing import *
 import sys
+import enum
 if sys.version_info < (3, 10):
     from typing_extensions import TypeAlias
 from abc import abstractmethod
@@ -311,6 +312,7 @@ class Signedness:
 class Signed(Signedness):
     @classmethod
     def convert(cls, bit_size: int, x: int) -> int:
+        assert bit_size > 0
         half = pow(2, bit_size - 1)
         if x < half:
             return x
@@ -442,6 +444,9 @@ class ContentBds(Content):
 
 
 class RuleContent:
+    """
+    Base class for different RuleContent possibilities.
+    """
 
     def __init__(self, bs: Bits) -> None:
         self.bs = bs
@@ -456,6 +461,10 @@ class RuleContent:
 
 
 class RuleContentContextFree(RuleContent):
+    """
+    A content structure that is fully defined in the specification
+    and is statically known.
+    """
     cv_content: ClassVar[Type[Content]]
 
     @classmethod
@@ -467,6 +476,11 @@ class RuleContentContextFree(RuleContent):
 
 
 class RuleContentDependent(RuleContent):
+    """
+    A content structure where all possible structures are defined
+    in the specification, but where the actual concrete structure depends
+    on a value of some other item, which is only known at run-time.
+    """
     cv_depends_on: ClassVar[List[List[str]]]
     cv_default_content: ClassVar[Type[Content]]
     cv_cases: ClassVar[List[Tuple[List[int], Type[Content]]]]
@@ -496,6 +510,11 @@ class RuleContentDependent(RuleContent):
 
 
 class Variation:
+    """
+    Base class (see subclasses below).
+    Subclasses follow asterix structure as defined here:
+    <https://zoranbosnjak.github.io/asterix-specs/struct.html>
+    """
     @classmethod
     @abstractmethod
     def parse(cls, bs: Bits) -> Union[ValueError, Tuple['Variation', Bits]]:
@@ -507,6 +526,9 @@ class Variation:
 
 
 class ItemBase:
+    """
+    Base class for Spare or NonSpare item (see subclasses below).
+    """
     @classmethod
     @abstractmethod
     def parse(cls, bs: Bits) -> Union[ValueError, Tuple['ItemBase', Bits]]:
@@ -518,6 +540,9 @@ class ItemBase:
 
 
 class RuleVariation:
+    """
+    Base class for different RuleVariation possibilities.
+    """
 
     @classmethod
     @abstractmethod
@@ -908,23 +933,33 @@ class Explicit(Variation):
         return self.bs.drop(8).to_bytes()
 
 
-def parse_fspec(
-        max_bytes: int, bs: Bits) -> Union[ValueError, Tuple[List[bool], Bits]]:
-    cnt = 0
-    flags: List[bool] = []
-    remaining = bs
-    while True:
-        if cnt >= max_bytes:
-            return ValueError('fspec max bytes exceeded')
-        cnt += 1
-        if len(remaining) < 8:
+class Fspec:
+    def __init__(self, bs: Bits):
+        self.bs = bs
+
+    @classmethod
+    def parse(cls, s: Bits) -> Union[ValueError, Tuple['Fspec', Bits]]:
+        n = 0
+        for i in s.to_bytes():
+            n += 8
+            if not (i & 0x01):
+                break
+        # empty input is not valid
+        if not n:
             return ValueError('overflow')
-        a, remaining = remaining.split_at(7)
-        flags.extend(list(a))
-        fx, remaining = remaining.head_tail()
-        if not fx:
-            break
-    return (flags, remaining)
+        # 'i' contains the last byte, which must have FX=0
+        if (i & 0x01):
+            return ValueError('overflow')
+        (a, b) = s.split_at(n)
+        return (cls(a), b)
+
+    def __iter__(self) -> Iterator[bool]:
+        n = 0
+        for flag in self.bs:
+            if n < 7:
+                yield flag
+            n += 1
+            n %= 8
 
 def unparse_fspec(items_list: List[Optional[Type[NonSpare]]],
                   args: Dict[str, Any]) -> Tuple[Bits, List[Any]]:
@@ -970,12 +1005,12 @@ class Compound(Variation):
 
     @classmethod
     def parse(cls, bs: Bits) -> Union[ValueError, Tuple['Compound', Bits]]:
-        result = parse_fspec(cls.cv_fspec_max_bytes, bs)
+        result = Fspec.parse(bs)
         if isinstance(result, ValueError):
             return result
-        flags, remaining = result
+        fspec, remaining = result
         items = {}
-        for (flag, nsp) in zip(flags, cls.cv_items_list):
+        for (flag, nsp) in zip(fspec, cls.cv_items_list):
             if not flag:
                 continue
             if nsp is None:
@@ -1019,6 +1054,10 @@ class Compound(Variation):
 
 
 class RuleVariationContextFree(RuleVariation):
+    """
+    Wrapper around Variation which is fully defined in the specification
+    and is statically known.
+    """
     cv_variation: ClassVar[Type[Variation]]
 
     def __init__(self, arg: Variation) -> None:
@@ -1048,6 +1087,11 @@ class RuleVariationContextFree(RuleVariation):
 
 
 class RuleVariationDependent(RuleVariation):
+    """
+    Wrapper around Variation, where all possible structures are defined
+    in the specification, but where the actual concrete structure depends
+    on a value of some other item, which is only known at run-time.
+    """
     cv_depends_on: ClassVar[List[List[str]]]
     cv_default_variation: ClassVar[Type[Variation]]
     cv_cases: ClassVar[List[Tuple[List[int], Type[Variation]]]]
@@ -1108,6 +1152,9 @@ def get_frn(items_list: List[Type[UapItemBase]], name: str) -> int:
         frn += 1
     raise ValueError(name, 'not found')
 
+class ParsingMode(enum.Enum):
+    StrictParsing = enum.auto()
+    PartialParsing = enum.auto()
 
 class Record:
     cv_fspec_max_bytes: ClassVar[int]
@@ -1119,32 +1166,55 @@ class Record:
         return cls.cv_items_dict[key]
 
     def __init__(self, bs: Bits, items1: Dict[str, NonSpare],
-                 items2: List[Tuple[str, NonSpare]]):
+                 items2: List[Optional[List[Tuple[str, NonSpare]]]]):
         self.bs = bs
         self.items_regular = items1
         self.items_rfs = items2
 
     @classmethod
-    def _parse(cls, bs: Bits) -> Union[ValueError, Tuple['Record', Bits]]:
-        result = parse_fspec(cls.cv_fspec_max_bytes, bs)
+    def _parse(cls, pm: ParsingMode, bs: Bits) -> \
+        Union[ValueError, Tuple['Record', Bits]]:
+        """Remark:
+        This function is using a less known python feature,
+        where the 'else' keyword is included at the end of 'for' loop.
+        This pattern allows nesteed for loops to break the outer loop,
+        when the inner loop breaks.
+        """
+
+        # make sure that we can use a simple 'if StrictParsing... else...'
+        # in the rest of this function (that is: there are only 2 modes).
+        if TYPE_CHECKING:
+            match pm:
+                case ParsingMode.StrictParsing: pass
+                case ParsingMode.PartialParsing: pass
+                case _: assert_never(pm)
+
+        result = Fspec.parse(bs)
         if isinstance(result, ValueError):
             return result
-        flags, remaining = result
-        items_regular: Dict[str, NonSpare] = {}
-        items_rfs: List[Tuple[str, NonSpare]] = []
-        for (flag, i) in zip(flags, cls.cv_items_list):
+        fspec, remaining = result
+        items1: Dict[str, NonSpare] = {}
+        items2: List[Optional[List[Tuple[str, NonSpare]]]] = []
+        for (flag, i) in zip(fspec, cls.cv_items_list):
             if not flag:
                 continue
             if issubclass(i, UapItemSpare):
-                return ValueError('fx bit set for spare item')
+                if pm == ParsingMode.StrictParsing:
+                    return ValueError('fx bit set for spare item')
+                else: break
             elif issubclass(i, UapItemRFS):
+                items3: List[Tuple[str, NonSpare]] = []
                 if len(remaining) < 8:
-                    return ValueError('overflow')
+                    if pm == ParsingMode.StrictParsing:
+                        return ValueError('overflow')
+                    else: break
                 a, remaining = remaining.split_at(8)
                 n = a.to_uinteger()
                 for j in range(n):
                     if len(remaining) < 8:
-                        return ValueError('overflow')
+                        if pm == ParsingMode.StrictParsing:
+                            return ValueError('overflow')
+                        else: break
                     b, remaining = remaining.split_at(8)
                     frn = b.to_uinteger()
                     try:
@@ -1153,86 +1223,112 @@ class Record:
                         nsp = T.cv_non_spare  # type: ignore
                         name = nsp.cv_name
                     except (IndexError, AttributeError, AssertionError):
-                        return ValueError('invalid FRN')
+                        if pm == ParsingMode.StrictParsing:
+                            return ValueError('invalid FRN')
+                        else: break
                     result2 = nsp.parse(remaining)
                     if isinstance(result2, ValueError):
-                        return result2
+                        if pm == ParsingMode.StrictParsing:
+                            return result2
+                        else: break
                     obj, remaining = result2
-                    items_rfs.append((name, obj))
+                    items3.append((name, obj))
+                else:
+                    items2.append(items3)
+                    continue
+                break
             elif issubclass(i, UapItem):
                 nsp = i.cv_non_spare
                 result2 = nsp.parse(remaining)
                 if isinstance(result2, ValueError):
-                    return result2
+                    if pm == ParsingMode.StrictParsing:
+                        return result2
+                    else: break
                 obj, remaining = result2
-                items_regular[nsp.cv_name] = obj
+                items1[nsp.cv_name] = obj
             else:
                 raise Exception('Unexpected', i)
         n = len(bs)
         m = len(remaining)
-        return (cls(bs.take(n - m), items_regular, items_rfs), remaining)
+
+        # In case of StrictParsing, we can grab the original bits,
+        # but for the PartialParsing, we need to recreate the record
+        # from parsed items, since the fspec might be different and
+        # the original bits are not exactly the same.
+        if pm == ParsingMode.StrictParsing:
+            r = cls(bs.take(n - m), items1, items2)
+        else:
+            r = cls._create(items1, items2)
+        return (r, remaining)
 
     @classmethod
-    def _create(cls, args: Dict[str, Any],
-                rfs: Optional[List[Any]] = None) -> Any:
-        bs = Bits.from_bytes(b'')
-        bs_rfs = Bits.from_uinteger(len(rfs or []), 0, 8)
-        items: List[Optional[Tuple[str, NonSpare]]] = []
-        items_regular: Dict[str, NonSpare] = {}
-        items_rfs: List[Tuple[str, NonSpare]] = []
+    def _create(cls, args: Dict[str, Any], *rfs_args: Any) -> Any:
+        # remove empty rfs elements from the tail of the rfss list
+        rfss = list(dropwhile(lambda x: not x, reversed(rfs_args)))
 
-        # process rfs items first
-        for (name, val) in (rfs or []):
-            frn = get_frn(cls.cv_items_list, name)
-            T = cls.cv_items_dict[name]
-            obj = T.create(val)  # type: ignore
-            items_rfs.append((name, obj))
-            bs_rfs += Bits.from_uinteger(frn, 0, 8)
-            bs_rfs += obj.unparse()
+        bs1 = Bits.from_bytes(b'')  # fspec
+        bs2 = Bits.from_bytes(b'')  # items
+        items1: Dict[str, NonSpare] = {}
+        items2: List[Optional[List[Tuple[str, NonSpare]]]] = []
+
+        def handle_rfs(rfs: List[Tuple[str, NonSpare]]) -> Any:
+            items = []
+            bs = Bits.from_uinteger(len(rfs), 0, 8)
+            for (name, val) in rfs:
+                frn = get_frn(cls.cv_items_list, name)
+                T = cls.cv_items_dict[name]
+                obj = T.create(val)  # type: ignore
+                items.append((name, obj))
+                bs += Bits.from_uinteger(frn, 0, 8)
+                bs += obj.unparse()
+            return (items, bs)
 
-        # process regular items
         cnt = 0
-        rfs_handled = rfs is None
         for i in cls.cv_items_list:
-            if (not args) and rfs_handled:
+            if (not args) and (not rfss):
                 break
             if cnt % 8 == 7:
                 cnt += 1
-                bs = bs.append_bit(True)  # FX
+                bs1 = bs1.append_bit(True)  # FX
             cnt += 1
             if issubclass(i, UapItem):
                 nsp = i.cv_non_spare
                 name = nsp.cv_name
                 try:
                     obj = nsp.create(args.pop(name))  # type: ignore
-                    bs = bs.append_bit(True)
-                    items.append((name, obj))
+                    bs1 = bs1.append_bit(True)
+                    bs2 += obj.unparse()
+                    items1[name] = obj
                 except KeyError:
-                    bs = bs.append_bit(False)
+                    bs1 = bs1.append_bit(False)
             elif issubclass(i, UapItemSpare):
-                bs = bs.append_bit(False)
+                bs1 = bs1.append_bit(False)
             elif issubclass(i, UapItemRFS):
-                rfs_handled = True
-                if rfs is None:
-                    bs = bs.append_bit(False)
+                if not rfss:
+                    bs1 = bs1.append_bit(False)
+                    items2.append(None)
                 else:
-                    bs = bs.append_bit(True)
-                    items.append(None)  # indicator for RFS fields
+                    rfs = rfss.pop(-1) # the list is reversed, pop from the tail
+                    if rfs is None:
+                        bs1 = bs1.append_bit(False)
+                    else:
+                        bs1 = bs1.append_bit(True)
+                        rfs_objs, bs = handle_rfs(rfs)
+                        items2.append(rfs_objs)
+                        bs2 += bs
             else:
                 raise ValueError('unexpected type', i)
-        if args:
+
+        # at this point, there should be no more items to process
+        if args or rfss:
             raise ValueError('items error', args)
+
+        # pad fspec
         while (cnt % 8):
             cnt += 1
-            bs = bs.append_bit(False)
-        for x in items:
-            if x is None:
-                bs += bs_rfs
-            else:
-                name, obj = x
-                bs += obj.unparse()
-                items_regular[name] = obj
-        return cls(bs, items_regular, items_rfs)
+            bs1 = bs1.append_bit(False)
+
+        return cls(bs1+bs2, items1, items2)
 
     def unparse(self) -> Bits:
         return self.bs
@@ -1242,9 +1338,12 @@ class Record:
 
     def _get_rfs_item(self, key: Any) -> Any:
         result = []
-        for (name, obj) in self.items_rfs:
-            if name == key:
-                result.append(obj)
+        for lst in self.items_rfs:
+            if lst is None:
+                continue
+            for (name, obj) in lst:
+                if name == key:
+                    result.append(obj)
         return result
 
     def _set_item(self, key: Any, val: Any) -> Any:
@@ -1270,7 +1369,7 @@ class Uap:
         remaining = bs
         rv = []
         while len(remaining):
-            result = r.parse(remaining)
+            result = r.parse(ParsingMode.StrictParsing, remaining)
             if isinstance(result, ValueError):
                 return result
             r, remaining = result
@@ -1316,7 +1415,7 @@ class UapMultiple(Uap):
             if not len(s):
                 yield acc
             for cls in lst:
-                result = cls._parse(s)
+                result = cls._parse(ParsingMode.StrictParsing, s)
                 if isinstance(result, ValueError):
                     continue
                 x, remaining = result
@@ -1363,10 +1462,11 @@ class Expansion:
             flags_bits, remaining = bs.split_at(n)
             flags = list(flags_bits)
         elif a == FspecFx:
-            result1 = parse_fspec(b, bs)
+            result1 = Fspec.parse(bs)
             if isinstance(result1, ValueError):
                 return result1
-            flags, remaining = result1
+            fspec, remaining = result1
+            flags = list(fspec)
         else:
             raise Exception('Unexpected cv_type', a)
         items = {}
