@@ -386,7 +386,7 @@ def test_explicit3a() -> None:
         '010': 0x0102,
         '072': re.unparse().to_bytes(),
     })
-    result = Cat.cv_record.parse(r.unparse())
+    result = Cat.cv_record.parse(ParsingMode.StrictParsing, r.unparse())
     assert not isinstance(result, ValueError)
     (r2, bs2) = result
     assert bs2.null()
@@ -421,7 +421,7 @@ def test_explicit3b() -> None:
         '010': 0x0102,
         '072': re.unparse().to_bytes(),
     })
-    result = Cat.cv_record.parse(r.unparse())
+    result = Cat.cv_record.parse(ParsingMode.StrictParsing, r.unparse())
     assert not isinstance(result, ValueError)
     (r2, bs2) = result
     assert bs2.null()
@@ -455,7 +455,7 @@ def test_explicit3c() -> None:
         '010': 0x0102,
         '072': re.unparse().to_bytes(),
     })
-    result = Cat.cv_record.parse(r.unparse())
+    result = Cat.cv_record.parse(ParsingMode.StrictParsing, r.unparse())
     assert not isinstance(result, ValueError)
     (r2, bs2) = result
     assert bs2.null()
@@ -575,7 +575,8 @@ def test_record() -> None:
         ('054', (1, 2)),
         ('061', [0xFF]),
     ])
-    assert with_rfs.unparse() == Bits.from_bytes(unhexlify(
+    bs = with_rfs.unparse()
+    assert bs == Bits.from_bytes(unhexlify(
         '410108030602AA02550112340A03040B03040C01FF'))
 
     assert with_rfs.get_item('000') is not None
@@ -586,6 +587,74 @@ def test_record() -> None:
     assert len(with_rfs.get_rfs_item('020')) == 0
     assert len(with_rfs.get_rfs_item('053')) == 1
 
+    # Check parsing (with RFS)
+    result = T.parse(ParsingMode.StrictParsing, bs)
+    assert not isinstance(result, ValueError)
+    (with_rfs2, bs2) = result
+    assert bs2.null()
+    assert with_rfs2.unparse() == bs
+
+def test_record_multiple_rfs() -> None:
+    T = Cat_003_1_0.cv_record
+
+    def check(r: Any, expected: str) -> None:
+        bs = r.unparse()
+        assert hexlify(bs.to_bytes()) == \
+            hexlify(Bits.from_bytes(unhexlify(expected)).to_bytes())
+        result = T.parse(ParsingMode.StrictParsing, bs)
+        assert not isinstance(result, ValueError)
+        (r2, bs2) = result
+        assert bs2.null()
+        assert r2.unparse() == bs
+
+    r0 = T.create({
+        '010': 0x01,
+    })
+    check(r0, '8001')
+
+    r1 = T.create({
+        '010': 0x01,
+    }, [
+        ('101', 0xAA),
+        ('102', 0x55),
+       ],
+    )
+    check(r1, '90010202aa0355')
+
+    r2 = T.create({
+        '010': 0x01,
+    }, [
+        ('101', 0xAA),
+        ('102', 0x55),
+       ],
+       [
+        ('201', 0xF1),
+        ('202', 0xF2),
+       ],
+    )
+    check(r2, '94010202aa0355020500f10700f2')
+
+    r3 = T.create({
+        '010': 0x01,
+    }, None,
+       [
+        ('201', 0xF1),
+        ('202', 0xF2),
+       ],
+    )
+    check(r3, '8401020500f10700f2')
+
+    r4 = T.create({
+        '010': 0x01,
+    }, [
+       ],
+       [
+        ('201', 0xF1),
+        ('202', 0xF2),
+       ],
+    )
+    check(r4, '940100020500f10700f2')
+
 
 def test_record_set_item() -> None:
     T = Cat_000_1_0.cv_record
@@ -595,7 +664,7 @@ def test_record_set_item() -> None:
     r1 = T.create({
         '000': 0x03,
     })
-    assert r1.unparse() == r0.unparse()
+    assert hexlify(r1.unparse().to_bytes()) == hexlify(r0.unparse().to_bytes())
 
 
 def test_record_del_item() -> None:
@@ -609,7 +678,7 @@ def test_record_del_item() -> None:
     r1 = T.create({
         '000': 0x03,
     })
-    assert r1.unparse() == r0.unparse()
+    assert hexlify(r1.unparse().to_bytes()) == hexlify(r0.unparse().to_bytes())
 
 
 def test_multiple_uaps() -> None:
@@ -656,7 +725,7 @@ def test_parse1() -> None:
         '410108030602AA02550112340A03040B03040C01FF'
     ]:
         bs = Bits.from_bytes(unhexlify(sample))
-        result = T.parse(bs)
+        result = T.parse(ParsingMode.StrictParsing, bs)
         assert not isinstance(result, ValueError)
         (r, bs2) = result
         assert bs2.null()
@@ -854,3 +923,49 @@ def test_parse5() -> None:
     results_track = Cat1.cv_uap.parse_any_uap(bs_track)
     assert len(results_track) == 1
 
+def test_parse_nonblocking() -> None:
+    """We encode a single record with a new edition, where some items
+    are added to the spec. We try to parse it as an old edition.
+    """
+
+    Ed0 = Cat_000_1_0
+    Ed1 = Cat_000_1_1
+
+    # old edition record
+    record0 = Ed0.cv_record.create({
+        '000': 0x00,
+        '010': 0x0102,
+    })
+
+    # new edition record (added item '200')
+    record1 = Ed1.cv_record.create({
+        '000': 0x00,
+        '010': 0x0102,
+        '200': 0xff,
+    })
+
+    # encode new record
+    s: Bits = record1.unparse()
+
+    # try to parse as a list with both editions
+    assert isinstance(Ed0.cv_uap.parse(s), ValueError)
+    assert not isinstance(Ed1.cv_uap.parse(s), ValueError)
+
+    # single record parsing combinations
+
+    # parsing with new edition shall not fail, regardless of parsing mode
+    for mode in ParsingMode:
+        result1 = Ed1.cv_record.parse(mode, s)
+        assert not isinstance(result1, ValueError)
+        r1, remaining = result1
+        assert remaining.null()
+        assert hexlify(record1.unparse().to_bytes()) == \
+                hexlify(r1.unparse().to_bytes())
+
+    # old edition, partial parsing shall not fail
+    result2 = Ed0.cv_record.parse(ParsingMode.PartialParsing, s)
+    assert not isinstance(result2, ValueError)
+    r2, remaining = result2
+    assert hexlify(record0.unparse().to_bytes()) == \
+            hexlify(r2.unparse().to_bytes())
+    assert remaining == Bits.from_uinteger(0xff, 0, 8)
